#ifndef lint
static char Rcs_Id[] =
	"$Header$";
#endif

/*
 * $Log$
 * Revision 2.8  1998/07/30 20:42:25  geoff
 * Add working BSD-socket support.  (Actually, there may still be a bug,
 * but I've never been able to track it.)
 *
 * Revision 2.7  1995/05/11  05:05:24  geoff
 * Attempt to make socket fifos work.  I don't think they do yet.
 *
 * Revision 2.6  87/03/31  22:32:39  geoff
 * Register-ize some strategic variables.
 * 
 * Revision 2.5  87/01/10  19:15:26  geoff
 * Add (untested) support for BSDSOCKET fifo's.
 * 
 * Revision 2.4  86/11/29  19:51:14  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.3  86/04/02  00:01:34  geoff
 * Switch from cron.h to config.h.
 * 
 * Revision 2.2  85/11/24  16:12:35  geoff
 * Fix some lint complaints.
 * 
 * Revision 2.1  85/10/27  17:31:45  geoff
 * Clean up some comments
 * 
 * Revision 1.1  85/10/13  21:43:57  geoff
 * Initial revision
 * 
 */

/*
 * UNIX-version-independent nonblocking FIFO operations.  These can be
 * compiled on any UNIX and they will work, albeit slowly on V7.
 *
 * Compilation options:  define one of the following three symbols
 * to select a fifo implementation:
 *
 * USGFIFO	To use USG named pipes (FIFO's) or an equivalent.
 * BSDSOCKET	To use BSD sockets or an equivalent.
 * VANILLA	If neither of the choices above is available.  This option
 *		implements FIFO's in the UNIX file system.  Warning:  this
 *		implementation is extremely inefficient.
 */

#include "config.h"		/* Only to get one of the above three guys */

/*
 * Interface descriptions:
 *
 * status = mkfifo ((char *) path, (int) mode);
 *
 *	A FIFO is created with the specified path name and protection mode.
 *	The return value is 0 on success, -1 on failure.  The path must not
 *	currently exist.
 *
 * fd = openfifo ((char *) path, (int) mode);
 *
 *	The specified FIFO is opened and a file descriptor is returned for
 *	access.  Only modes of 0 (O_RDONLY) or 1 (O_WRONLY) are guaranteed
 *	to be supported.  O_RDONLY opens never delay (this is important),
 *	but an O_WRONLY open MAY delay on some systems until a process
 *	opens the fifo for read.
 *
 * nbytes = readfifo (fd, buf, count);
 *
 *	Up to "count" bytes are read from the fifo and placed in "buf".
 *	If no bytes are available, returns zero.
 *
 * nbytes = writefifo (fd, buf, count);
 *
 *	"Count" bytes are written to the fifo from "buf".  The number of
 *	bytes actually written should match "count" unless there is an error.
 *
 * status = closefifo (fd);
 *
 *	The specified fifo is closed.  For O_WRONLY opens, any
 *	previously-written information is preserved.  For O_RDONLY opens,
 *	the disposition of unread information is implementation-dependent.
 *
 * status = rmfifo (path);
 *
 *	Removes (unlinks) a fifo.  Any process attempting to use the fifo
 *	after this (even if it is already open) MAY receive an error on
 *	some systems.
 */

int	mkfifo ();		/* Make a FIFO */
int	openfifo ();		/* Open an existing FIFO */
int	readfifo ();		/* Read a FIFO */
int	writefifo ();		/* Write a FIFO */
int	closefifo ();		/* Close a FIFO */
int	rmfifo ();		/* Remove a FIFO */

#ifdef USGFIFO
#include <fcntl.h>

int mkfifo (path, mode)		/* Make a FIFO */
    char *		path;	/* Name to give it */
    int			mode;	/* Protection modes */
    {
    return mknod (path, mode | 010000, 0);
    }

int openfifo (path, mode)	/* Open an existing FIFO */
    char *		path;	/* Name of the FIFO */
    int			mode;	/* Open mode (read or write) */
    {
    return open (path, (mode) ? (mode) : (O_RDONLY | O_NDELAY));
    }

int readfifo (fd, buf, size)	/* Read a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to read into */
    int			size;	/* Number of characters to read */
    {
    return read (fd, buf, (unsigned) size);
    }

int writefifo (fd, buf, size)	/* Write a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to write from */
    int			size;	/* Number of characters to write */
    {
    return write (fd, buf, (unsigned) size);
    }
    
int closefifo (fd)		/* Close a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    {
    return close (fd);
    }

int rmfifo (path)		/* Remove a FIFO */
    char *		path;	/* Name of the FIFO to remove */
    {
    return unlink (path);
    }
#endif

#ifdef BSDSOCKET
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>

int mkfifo (path, mode)		/* Make a FIFO */
    char *		path;	/* Name to give it */
    int			mode;	/* Protection modes */
    {
    /*
     * In BSD, a FIFO is created by opening it for read.  However, the
     * creation will only work if the fifo does not already exist, so we must
     * unlink it first.
     */
    if (unlink (path) < 0  &&  errno != ESRCH)
	return -1;
    return 0;
    }

int openfifo (path, mode)	/* Open an existing FIFO */
    char *		path;	/* Name of the FIFO */
    int			mode;	/* Open mode (read or write) */
    {
    struct sockaddr	addr;	/* Address to give to the socket */
#if !defined(O_NONBLOCK) || defined(ultrix) || defined(hpux) || defined(AIXV3)
#if defined(FIOSNBIO)  ||  (defined(AIXV3)  &&  defined(FIONBIO))
    int			blockarg; /* Used for setting nonblocking socket */
#endif
#endif
    reg1 int		fd;	/* FD we will return to the user */
    struct linger	lingerbuf; /* Buffer for turning on SO_LINGER */
    int			oumask;	/* Process umask value */
    reg2 int		saverr;	/* Room to save errno in case of failures */

    addr.sa_family = AF_UNIX;
    strncpy (addr.sa_data, path, sizeof (addr.sa_data));
    addr.sa_data[sizeof (addr.sa_data) - 1] = '\0';
    if (mode == 0)		/* Open for read:  user is a listener */
	{
	fd = socket (AF_UNIX, SOCK_STREAM, 0);
	/* If the result is ADDRINUSE, try unlinking it */
	if (fd < 0  &&  errno == EADDRINUSE)
	    {
	    if (unlink (path) < 0)
		return -1;
	    fd = socket (AF_UNIX, SOCK_STREAM, 0);
	    }
	if (fd < 0)
	    return -1;
	if (bind (fd, &addr, sizeof (addr)) < 0)
	    {
	    /* If the result is ADDRINUSE, try unlinking it */
	    if (errno == EADDRINUSE)
		{
		if (unlink (path) < 0)
		    return -1;
		fd = socket (AF_UNIX, SOCK_STREAM, 0);
		if (fd < 0)
		    return -1;
		if (bind (fd, &addr, sizeof (addr)) < 0)
		    {
		    saverr = errno;
		    (void) close (fd);
		    errno = saverr;
		    return -1;
		    }
		}
	    else
		{
		saverr = errno;
		(void) close (fd);
		errno = saverr;
 		return -1;
		}
	    }
	oumask = umask (0);
	(void) umask (oumask);
	(void) fchmod (fd, 0666 & ~oumask);
 	if (listen (fd, 1) < 0)	/* Ask to hear about client requests */
	    {
	    saverr = errno;
	    (void) close (fd);
	    errno = saverr;
	    return -1;
	    }
	/*
	 * Ask for the socket to be nonblocking on accept requests.
	 * This code is stolen (indirectly) from the X server.
	 */
#if defined(O_NONBLOCK) && !defined(ultrix) && !defined(hpux) && !defined(AIXV3)
	if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0)
#else
#ifdef FIOSNBIO
	blockarg = 1;
	if (ioctl (fd, FIOSNBIO, &blockarg) < 0)
#else
#if defined(AIXV3) && defined(FIONBIO)
	blockarg = 1;
	if (ioctl(fd, FIONBIO, &blockarg) < 0)
#else
	if (fcntl (fd, F_SETFL, FNDELAY) < 0)
#endif
#endif
#endif
	    {
	    saverr = errno;
	    (void) close (fd);
	    errno = saverr;
	    return -1;
	    }
 	return fd;
 	}
    else			/* Open for write: user is a client */
 	{
 	fd = socket (AF_UNIX, SOCK_STREAM, 0);
 	if (fd < 0)
	    return fd;
 	if (connect (fd, &addr, sizeof (addr)) < 0)
	    {
	    saverr = errno;
	    (void) close (fd);
	    errno = saverr;
	    return -1;
	    }
	lingerbuf.l_onoff = 1;
	lingerbuf.l_linger = 0x7fffffff; /* Linger essentially forever */
	(void) setsockopt (fd, SOL_SOCKET, SO_LINGER, &lingerbuf,
	  sizeof lingerbuf);
 	return fd;
 	}
    }
    
int readfifo (fd, buf, size)	/* Read a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to read into */
    int			size;	/* Number of characters to read */
    {
    int			len;	/* Length of socket address */
    reg1 int		nbytes;	/* Number of bytes read returned */
    static int		opensock = -1; /* Socket to do I/O on */
    struct sockaddr	saddr;	/* Address of client that asked for I/O */
    reg2 int		saverr;	/* Room to save errno in case of trouble */

    while (opensock < 0)
	{
	len = sizeof (saddr);
	opensock = accept (fd, &saddr, &len); /* Get a client */
	if (opensock < 0)
	    {
	    if (errno == EWOULDBLOCK)
		return 0;
	    else
		return opensock;
	    }
	}
    nbytes = read (opensock, buf, size);
    if (nbytes <= 0)
	{
	saverr = errno;	/* Preserve error codes */
	(void) close (opensock);
	opensock = -1;
	errno = saverr;
	}
    return nbytes;
    }

int writefifo (fd, buf, size)	/* Write a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to write from */
    int			size;	/* Number of characters to write */
    {

    return write (fd, buf, size);
    }

int closefifo (fd)		/* Close a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    {

    return shutdown (fd, 1);
    }

int rmfifo (path)		/* Remove a FIFO */
    char *		path;	/* Name of the FIFO to remove */
    {
    return unlink (path);
    }
#endif

#ifdef VANILLA

#include <errno.h>
#include <sys/param.h>

static void	lockit ();	/* Lock the fifo data file */
static void	unlockit ();	/* Unlock the fifo data file */
static int	myrand ();	/* Simple random-number generator */

#ifndef NOFILE
#define NOFILE	20		/* Number of files a process can have */
#endif

#define ADDSIZE		6	/* Size of longest string added to path */
#define MAXSLEEP	60	/* Never sleep (much) longer than this */

static char *		fifopaths[NOFILE]; /* Pathnames of data files */
static char *		fifolocks[NOFILE]; /* Pathnames of lock files */
static long		fifopos[NOFILE]; /* Positions of data files */
static long		ranseed;	/* Seed for 'random' numbers */

extern long		lseek ();	/* Seek on file */
extern char *		malloc ();	/* Memory allocator */
extern long		time ();	/* Return system time */

int mkfifo (path, mode)		/* Make a FIFO */
    reg2 char *		path;	/* Name to give it */
    reg5 int		mode;	/* Protection modes */
    {
    reg3 int		fd;	/* File descriptor for FIFO control files */
    reg1 char *		buf;	/* Buffer for making file names */
    reg4 int		pathsize; /* Size of the path name */

    pathsize = strlen (path);
    buf = malloc (pathsize + 1 + ADDSIZE);
    if (buf == NULL)
	return -1;
    (void) strcpy (buf, "mkdir ");
    strcat (buf, path);
    if (system (buf))		/* Try to make the directory */
	return -1;
    chmod (path, mode | ((mode & 0444) >> 2)); /* Set up proper dir modes */
    (void) strcpy (buf, path);
    (void) strcpy (&buf[pathsize], "/data"); /* Make the data file */
    fd = creat (buf, mode);
    if (fd < 0)
	{			/* Couldn't make it - remove the directory */
	(void) strcpy (buf, "rmdir ");
	strcat (buf, path);
	system (buf);
	free (buf);
	return -1;
	}
    (void) close (fd);
    free (buf);
    ranseed = time ((long *) 0);
    return 0;
    }

int openfifo (path, mode)	/* Open an existing FIFO */
    char *		path;	/* Name of the FIFO */
    int			mode;	/* Open mode (read or write) */
    {
    char *		buf;	/* Buffer for FIFO path name */
    char *		lbuf;	/* Buffer for lock file path name */
    int			fd;	/* File descriptor for the opened fifo */

    buf = malloc (strlen (path) + 1 + ADDSIZE);
    if (buf == NULL)
	{
	errno = ENOMEM;
	return -1;
	}
    lbuf = malloc (strlen (path) + 1 + ADDSIZE);
    if (lbuf == NULL)
	{
	free (buf);
	errno = ENOMEM;
	return -1;
	}
    (void) strcpy (buf, path);
    strcat (buf, "/data");
    (void) strcpy (lbuf, path);
    strcat (lbuf, "/lock");
    fd = open (buf, mode);
    if (fd >= NOFILE)		/* We can't handle this */
	{
	(void) close (fd);
	fd = -1;
	}
    if (fd < 0)
	{
	free (lbuf);
	free (buf);
	return -1;
	}
    fifopaths[fd] = buf;
    fifolocks[fd] = lbuf;
    (void) lseek (fd, 0L, 2);
    fifopos[fd] = lseek (fd, 0L, 1);
    return fd;
    }

int readfifo (fd, buf, size)	/* Read a FIFO */
    reg1 int		fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to read into */
    int			size;	/* Number of characters to read */
    {
    int			saveerr; /* Space to save errno */

    lockit (fifopaths[fd], fifolocks[fd]);
    (void) lseek (fd, fifopos[fd], 0);
    size = read (fd, buf, size);
    saveerr = errno;
    fifopos[fd] = lseek (fd, 0L, 1);	/* Get current position */
    if (size <= 0)			/* Is fifo empty now? */
	{
	(void) close (creat (fifopaths[fd], 0666)); /* Truncate the fifo */
					/* The mode is unchanged */
	fifopos[fd] = 0;		/* Return to start of file */
	}
    unlockit (fifopaths[fd], fifolocks[fd]);
    errno = saveerr;
    return size;
    }

int writefifo (fd, buf, size)	/* Write a FIFO */
    int			fd;	/* File descriptor for the FIFO */
    char *		buf;	/* Buffer to write from */
    int			size;	/* Number of characters to write */
    {
    int			saveerr; /* Space to save errno */

    lockit (fifopaths[fd], fifolocks[fd]);
    (void) lseek (fd, 0L, 2);	/* Head out to EOF */
    size = write (fd, buf, size);
    saveerr = errno;
    unlockit (fifopaths[fd], fifolocks[fd]);
    errno = saveerr;
    return size;
    }

int closefifo (fd)		/* Close a FIFO */
    reg1 int		fd;	/* File descriptor for the FIFO */
    {
    free (fifopaths[fd]);
    free (fifolocks[fd]);
    fifopaths[fd] = NULL;
    fifolocks[fd] = NULL;
    return close (fd);
    }

int rmfifo (path)		/* Remove a FIFO */
    char *		path;	/* Name of the FIFO to remove */
    {
    reg1 char *		buf;	/* Buffer for building command */

    buf = malloc (strlen (path) + 1 + ADDSIZE);
    if (buf == NULL)
	return -1;
    (void) strcpy (buf, "rm -rf ");
    strcat (buf, path);
    if (system (buf))
	{
	free (buf);
	return -1;
	}
    free (buf);
    return 0;
    }

static void lockit (path, lock)	/* Lock 'path' using links to 'lock' */
    reg2 char *		path;	/* Path name to be locked */
    reg3 char *		lock;	/* Lock file to create */
    {
    reg1 int		sleepmask = 1;	/* Exponentially growing sleep time */

    while (link (path, lock) < 0  &&  errno == EEXIST)
	{
	ranseed = ranseed * 1001 + 123;	/* Multiplicative congruential */
	(void) sleep (sleepmask & (ranseed >> 8)); /* Random delay */
	if (sleepmask < MAXSLEEP) /* Increase sleep time */
	    sleepmask |= sleepmask << 1;
	}
    }

static void unlockit (path, lock) /* Unlock 'path' using 'lock' */
    char *		path;	/* Path name to be unlocked */
    char *		lock;	/* Path name of lock file */
    {
    (void) unlink (lock);	/* That wasn't hard, was it? */
    }
#endif
