#!/usr/bin/perl -w

#
# location_monitor.pl
# $Id: location_monitor.pl 3572 2003-02-05 05:52:28Z johnh $
#
# Copyright (C) 1999 by John Heidemann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#



sub usage {
    print STDERR <<END;
usage: $0 [-vDR]

monitor and filter location, sending signals to lcron if it changes

With no arguments, runs once.
Options:

-D puts it in "daemon" mode where it waits and runs every SIGIO
    and puts its pid in /var/run/location_monitor.pid
-R report its presence in /var/run/netreport, causing it to get SIGIOs
    on RedHat-Linux
-v verbose
-L LOGFILE   log to logfile

END
    exit 1;
}

use strict;

use Getopt::Long;
&usage if ($#ARGV >= 0 && $ARGV[0] eq '-?');
my(%opts);
&GetOptions(\%opts, qw(v D R L=s));

# These next values will change in installation
# If they're not correct, you probably didn't do make install.lcron.
my($BINDIR) = '/usr/local/lib/xcron/bin';
my($LIBDIR) = '/usr/local/lib/xcron/lib';

my($verbose) = defined($opts{'v'});
my($logfile) = "-";
if (defined($opts{'L'})) {
    $logfile = $opts{'L'};
    $verbose = 1;
    close(STDOUT);
    open(STDOUT, ">$logfile") || die "$0: cannot write to $logfile\n";
    $| = 1;  # make unbuffered
};

if (defined($opts{'D'})) {
    my($f) = fork;
    die "$0: fork failed\n" if (!defined($f));
    exit 0 if ($f != 0);
};

my($need_netreport) = defined($opts{'R'});
&netreport() if ($need_netreport);
&daemon_mode if (defined($opts{'D'}));
&check();
exit 0;


sub netreport {
    open (PIDFILE, ">/var/run/netreport/$$") || die "$0: cannot create /var/run/netreport/$$.\n";
    print PIDFILE "$0\n", scalar localtime, "\n";
    close PIDFILE;
}

sub save_pid {
    my($f) = "/var/run/location_monitor.pid";
    open (PIDFILE, ">$f") || die "$0: cannot create $f.\n";
    print PIDFILE "$$\n";
    close PIDFILE;
}


#
# check for a location change:
#    run netstat
#    filter the output through netstat_mapping
#    save the results in events
#
sub check {
    printf "location_monitor at " . (scalar localtime) . "\n" if ($verbose);

    # icky but portable: figure out where netstat is
    my($netstat);
    foreach (qw(/usr/bin/netstat /bin/netstat /etc/netstat /usr/ucb/netstat /dev/null)) {
	$netstat = $_, last if (-x $_);
    };
    die "$0: cannot find netstat\n" if ($netstat eq '/dev/null');

    my(@gws) = ();
    my($gw);
    open(NS, "$netstat -rn |") || die "$0: $netstat failed.\n";
    open(GW, ">$LIBDIR/events/gateways") || die "$0: cannot open '$LIBDIR/events/gateways'.\n";
    chmod(0644, "$LIBDIR/events/gateways");
    while (<NS>) {  
	my(@f) = split;
	my($addr);
	foreach $addr ($f[0], $f[1]) {
	    if ($addr !~ /^(0\.0\.0\.0|Gateway|IP)/) {
		print GW "$addr\n";
		push(@gws, $addr);
	    };
	};
    };
    close NS;
    close GW;

    my(%gw_to_place) = ();
    if (-f "$LIBDIR/netstat_mapping") {
	open(MAP, "<$LIBDIR/netstat_mapping") || warn "$0: cannot open '$LIBDIR/netstat_mapping'.\n";
	while (<MAP>) {
	    next if (/^\s*#/);  # comments
	    chomp;
	    my(@f) = split;
	    next if ($#f < 0);   # blank lines
	    # warn "$0: syntax errors in netstat_mapping, skipping: $_\n";
	    my($gw) = shift @f;
	    $gw_to_place{$gw} = [] if (!defined($gw_to_place{$gw}));
	    push (@{$gw_to_place{$gw}}, @f);
	};
	close MAP;
    };
    open(PL, ">$LIBDIR/events/places") || die "$0: cannot open '$LIBDIR/events/places'.\n";
    chmod(0644, "$LIBDIR/events/places");
    foreach $gw (@gws) {
	if (defined($gw_to_place{$gw})) {
	    my($pl);
	    foreach $pl (@{$gw_to_place{$gw}}) {
	    	print PL "$pl\n";
	    };
	};
    };
    close PL;

    system("$BINDIR/tellxcron", "-L");

    # put oursevles back for next round
    &netreport() if ($need_netreport);
}

sub daemon_exit_error {
    exit 1;
}

sub daemon_exit_noerror {
    exit 0;
}

#
# daemon mode is a redhat thing:
# The system will send you SIGIO when interfaces change.
#
sub daemon_mode {
    &save_pid();
    $SIG{'IO'} = \&check;
    $SIG{'HUP'} = \&daemon_exit_noerror;
    $SIG{'INT'} = \&daemon_exit_error;
    $SIG{'TERM'} = \&daemon_exit_error;

    # start by checking!
    &check();

    for (;;) {
	sleep 6000;
    };
    die "$0: error, daemon mode ecaped infinte loop.\n";
}

